import { EitherModule } from "./emscripten-types";
import { JSContextPointer, JSRuntimePointer, JSValuePointer, JSValueConstPointer, JSModuleDefPointer } from "./ffi-types";
import { EitherFFI } from "./types";
/**
 * @private
 */
export declare type CToHostCallbackFunctionImplementation = (ctx: JSContextPointer, this_ptr: JSValueConstPointer, argc: number, argv: JSValueConstPointer, fn_id: number) => JSValuePointer;
/**
 * @private
 */
export declare type CToHostAsyncCallbackFunctionImplementation = (ctx: JSContextPointer, this_ptr: JSValueConstPointer, argc: number, argv: JSValueConstPointer, fn_id: number) => JSValuePointer | Promise<JSValuePointer>;
/**
 * @private
 */
export declare type CToHostInterruptImplementation = (rt: JSRuntimePointer) => 0 | 1;
/**
 * @private
 */
export declare type CToHostModuleLoaderImplementation = (rt: JSRuntimePointer, ctx: JSContextPointer, moduleName: string) => JSModuleDefPointer | Promise<JSModuleDefPointer>;
/**
 * @private
 */
export interface ContextCallbacks {
    cToHostCallbackFunction: CToHostCallbackFunctionImplementation;
}
/**
 * @private
 */
export interface RuntimeCallbacks {
    cToHostInterrupt: CToHostInterruptImplementation;
    cToHostLoadModule: CToHostModuleLoaderImplementation;
}
/**
 * We use static functions per module to dispatch runtime or context calls from
 * C to the host.  This class manages the indirection from a specific runtime or
 * context pointer to the appropriate callback handler.
 *
 * @private
 */
export declare class QuickJSModuleCallbacks {
    private module;
    private ffi;
    private contextCallbacks;
    private runtimeCallbacks;
    constructor(module: EitherModule, ffi: EitherFFI);
    setRuntimeCallbacks(rt: JSRuntimePointer, callbacks: RuntimeCallbacks): void;
    deleteRuntime(rt: JSRuntimePointer): void;
    setContextCallbacks(ctx: JSContextPointer, callbacks: ContextCallbacks): void;
    deleteContext(ctx: JSContextPointer): void;
    private maybeSuspendOrResumeAsync;
    private cToHostCallbacks;
}
