"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const _1 = require(".");
const memory_1 = require("./memory");
const fs = __importStar(require("fs"));
function sleep(n) {
    return new Promise((resolve) => setTimeout(resolve, n));
}
async function main() {
    const QuickJS = await (0, _1.getQuickJS)();
    const ffi = QuickJS.ffi;
    const memory = new memory_1.ModuleMemory(QuickJS.module);
    const db = ffi.QTS_DB_New();
    if (!db) {
        throw new Error("db is null");
    }
    console.log("Created DB, pointer:", db);
    const exec = (sql) => {
        return memory.newHeapCharPointer(sql).consume((ptr) => {
            ffi.QTS_DB_Exec(db, ptr.value, 1);
        });
    };
    const createTable = async (index) => {
        const tableName = `map${index}`;
        exec(`CREATE TABLE ${tableName} (key TEXT PRIMARY KEY, value INTEGER)`);
        console.log(`Created table ${tableName}`);
        await sleep(5000);
        exec(`DROP TABLE ${tableName}`);
        console.log(`Dropped table ${tableName}`);
    };
    const statFile = fs.openSync("sqlite-mem.tsv", "w");
    fs.writeFileSync(statFile, "rss\theap_total\theap_used\n");
    for (let i = 0; i < 1000; i++) {
        console.log(i);
        if (i % 1 === 0) {
            console.log("iter", i);
            const stat = process.memoryUsage();
            const row = [stat.rss, stat.heapTotal, stat.heapUsed].map((num) => (num / 1024 / 1024).toFixed(2));
            fs.writeFileSync(statFile, row.join("\t") + "\n");
        }
        const promises = [];
        // 10000 tables: 1.4g
        // 10000 dbs: 1.6g
        for (let inner = 0; inner < 10000; inner++) {
            promises.push(createTable(inner));
        }
        await Promise.all(promises);
    }
    fs.closeSync(statFile);
}
main();
//# sourceMappingURL=sqlite_test.js.map