"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QuickJSVm = exports.PureQuickJSVm = void 0;
var deferred_promise_1 = require("./deferred-promise");
var lifetime_1 = require("./lifetime");
/**
 * @private
 */
var QuickJSVmMemory = /** @class */ (function () {
    function QuickJSVmMemory(args) {
        var _this = this;
        this.scope = new lifetime_1.Scope();
        this.copyJSValue = function (ptr) {
            return _this.ffi.QTS_DupValuePointer(_this.ctx.value, ptr);
        };
        this.freeJSValue = function (ptr) {
            _this.ffi.QTS_FreeValuePointer(_this.ctx.value, ptr);
        };
        this.owner = args.owner;
        this.module = args.module;
        this.ffi = args.ffi;
        this.rt = this.scope.manage(args.rt);
        this.ctx = this.scope.manage(args.ctx);
    }
    Object.defineProperty(QuickJSVmMemory.prototype, "alive", {
        get: function () {
            return this.scope.alive;
        },
        enumerable: false,
        configurable: true
    });
    QuickJSVmMemory.prototype.dispose = function () {
        return this.scope.dispose();
    };
    /**
     * Track `lifetime` so that it is disposed when this scope is disposed.
     */
    QuickJSVmMemory.prototype.manage = function (lifetime) {
        return this.scope.manage(lifetime);
    };
    QuickJSVmMemory.prototype.assertOwned = function (handle) {
        if (handle.owner && handle.owner !== this.owner) {
            throw new Error('Given handle created by a different VM');
        }
    };
    QuickJSVmMemory.prototype.heapValueHandle = function (ptr) {
        return new lifetime_1.Lifetime(ptr, this.copyJSValue, this.freeJSValue, this.owner);
    };
    QuickJSVmMemory.prototype.toPointerArray = function (handleArray) {
        var _this = this;
        var typedArray = new Int32Array(handleArray.map(function (handle) { return handle.value; }));
        var numBytes = typedArray.length * typedArray.BYTES_PER_ELEMENT;
        var ptr = this.module._malloc(numBytes);
        var heapBytes = new Uint8Array(this.module.HEAPU8.buffer, ptr, numBytes);
        heapBytes.set(new Uint8Array(typedArray.buffer));
        return new lifetime_1.Lifetime(ptr, undefined, function (ptr) { return _this.module._free(ptr); });
    };
    QuickJSVmMemory.prototype.newMutablePointerArray = function (length) {
        var _this = this;
        var zeros = new Int32Array(new Array(length).fill(0));
        var numBytes = zeros.length * zeros.BYTES_PER_ELEMENT;
        var ptr = this.module._malloc(numBytes);
        var typedArray = new Int32Array(this.module.HEAPU8.buffer, ptr, length);
        typedArray.set(zeros);
        return new lifetime_1.Lifetime({ typedArray: typedArray, ptr: ptr }, undefined, function (value) { return _this.module._free(value.ptr); });
    };
    QuickJSVmMemory.prototype.newHeapCharPointer = function (string) {
        var _this = this;
        var numBytes = this.module.lengthBytesUTF8(string) + 1;
        var ptr = this.module._malloc(numBytes);
        this.module.stringToUTF8(string, ptr, numBytes);
        return new lifetime_1.Lifetime(ptr, undefined, function (value) { return _this.module._free(value); });
    };
    return QuickJSVmMemory;
}());
/**
 * Implements methods that are always synchronous - such methods should never
 * call into a function that suspends when built with ASYNCIFY.
 *
 * @private
 */
var PureQuickJSVm = /** @class */ (function () {
    /**
     * Use {@link QuickJS.createVm} to create a QuickJSVm instance.
     */
    function PureQuickJSVm(args) {
        var _this = this;
        /** @private */
        this._undefined = undefined;
        /** @private */
        this._null = undefined;
        /** @private */
        this._false = undefined;
        /** @private */
        this._true = undefined;
        /** @private */
        this._global = undefined;
        /**
         * @hidden
         */
        this.cToHostInterrupt = function (rt) {
            if (rt !== _this.rt.value) {
                throw new Error('QuickJSVm instance received C -> JS interrupt with mismatched rt');
            }
            var fn = _this.interruptHandler;
            if (!fn) {
                throw new Error('QuickJSVm had no interrupt handler');
            }
            return fn(_this.owner) ? 1 : 0;
        };
        this.owner = this;
        this.module = args.module;
        this.ffi = args.ffi;
        this.rt = args.rt;
        this.ctx = args.ctx;
        this.memory = new QuickJSVmMemory(__assign(__assign({}, args), { owner: this.owner }));
    }
    Object.defineProperty(PureQuickJSVm.prototype, "alive", {
        // @implement Disposable ----------------------------------------------------
        get: function () {
            return this.memory.alive;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Dispose of this VM's underlying resources.
     *
     * @throws Calling this method without disposing of all created handles
     * will result in an error.
     */
    PureQuickJSVm.prototype.dispose = function () {
        this.memory.dispose();
    };
    Object.defineProperty(PureQuickJSVm.prototype, "undefined", {
        // Globals ------------------------------------------------------------------
        /**
         * [`undefined`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/undefined).
         */
        get: function () {
            if (this._undefined) {
                return this._undefined;
            }
            // Undefined is a constant, immutable value in QuickJS.
            var ptr = this.ffi.QTS_GetUndefined();
            return (this._undefined = new lifetime_1.StaticLifetime(ptr));
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(PureQuickJSVm.prototype, "null", {
        /**
         * [`null`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/null).
         */
        get: function () {
            if (this._null) {
                return this._null;
            }
            // Null is a constant, immutable value in QuickJS.
            var ptr = this.ffi.QTS_GetNull();
            return (this._null = new lifetime_1.StaticLifetime(ptr));
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(PureQuickJSVm.prototype, "true", {
        /**
         * [`true`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/true).
         */
        get: function () {
            if (this._true) {
                return this._true;
            }
            // True is a constant, immutable value in QuickJS.
            var ptr = this.ffi.QTS_GetTrue();
            return (this._true = new lifetime_1.StaticLifetime(ptr));
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(PureQuickJSVm.prototype, "false", {
        /**
         * [`false`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/false).
         */
        get: function () {
            if (this._false) {
                return this._false;
            }
            // False is a constant, immutable value in QuickJS.
            var ptr = this.ffi.QTS_GetFalse();
            return (this._false = new lifetime_1.StaticLifetime(ptr));
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(PureQuickJSVm.prototype, "global", {
        /**
         * [`global`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects).
         * A handle to the global object inside the interpreter.
         * You can set properties to create global variables.
         */
        get: function () {
            if (this._global) {
                return this._global;
            }
            // The global is a JSValue, but since it's lifetime is as long as the VM's,
            // we should manage it.
            var ptr = this.ffi.QTS_GetGlobalObject(this.ctx.value);
            // Automatically clean up this reference when we dispose(
            this.memory.manage(this.memory.heapValueHandle(ptr));
            // This isn't technically a static lifetime, but since it has the same
            // lifetime as the VM, it's okay to fake one since when the VM is
            // disposed, no other functions will accept the value.
            this._global = new lifetime_1.StaticLifetime(ptr, this.owner);
            return this._global;
        },
        enumerable: false,
        configurable: true
    });
    // New values ---------------------------------------------------------------
    /**
     * Converts a Javascript number into a QuickJS value.
     */
    PureQuickJSVm.prototype.newNumber = function (num) {
        return this.memory.heapValueHandle(this.ffi.QTS_NewFloat64(this.ctx.value, num));
    };
    /**
     * Create a QuickJS [string](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String) value.
     */
    PureQuickJSVm.prototype.newString = function (str) {
        var _this = this;
        var ptr = this.memory.newHeapCharPointer(str).consume(function (charHandle) {
            return _this.ffi.QTS_NewString(_this.ctx.value, charHandle.value);
        });
        return this.memory.heapValueHandle(ptr);
    };
    /**
     * `{}`.
     * Create a new QuickJS [object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Object_initializer).
     *
     * @param prototype - Like [`Object.create`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/create).
     */
    PureQuickJSVm.prototype.newObject = function (prototype) {
        if (prototype) {
            this.memory.assertOwned(prototype);
        }
        var ptr = prototype
            ? this.ffi.QTS_NewObjectProto(this.ctx.value, prototype.value)
            : this.ffi.QTS_NewObject(this.ctx.value);
        return this.memory.heapValueHandle(ptr);
    };
    /**
     * `[]`.
     * Create a new QuickJS [array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array).
     */
    PureQuickJSVm.prototype.newArray = function () {
        var ptr = this.ffi.QTS_NewArray(this.ctx.value);
        return this.memory.heapValueHandle(ptr);
    };
    /**
     * Create a new [[QuickJSDeferredPromise]]. Use `deferred.resolve(handle)` and
     * `deferred.reject(handle)` to fulfill the promise handle available at `deferred.handle`.
     * Note that you are responsible for calling `deferred.dispose()` to free the underlying
     * resources; see the documentation on [[QuickJSDeferredPromise]] for details.
     */
    PureQuickJSVm.prototype.newPromise = function () {
        var _this = this;
        return lifetime_1.Scope.withScope(function (scope) {
            var mutablePointerArray = scope.manage(_this.memory.newMutablePointerArray(2));
            var promisePtr = _this.ffi.QTS_NewPromiseCapability(_this.ctx.value, mutablePointerArray.value.ptr);
            var promiseHandle = _this.memory.heapValueHandle(promisePtr);
            var _a = Array.from(mutablePointerArray.value.typedArray).map(function (jsvaluePtr) { return _this.memory.heapValueHandle(jsvaluePtr); }), resolveHandle = _a[0], rejectHandle = _a[1];
            return new deferred_promise_1.QuickJSDeferredPromise({
                owner: _this.owner,
                promiseHandle: promiseHandle,
                resolveHandle: resolveHandle,
                rejectHandle: rejectHandle,
            });
        });
    };
    // Runtime management -------------------------------------------------------
    /**
     * In QuickJS, promises and async functions create pendingJobs. These do not execute
     * immediately and need to be run by calling [[executePendingJobs]].
     *
     * @return true if there is at least one pendingJob queued up.
     */
    PureQuickJSVm.prototype.hasPendingJob = function () {
        return Boolean(this.ffi.QTS_IsJobPending(this.rt.value));
    };
    /**
     * Set a callback which is regularly called by the QuickJS engine when it is
     * executing code. This callback can be used to implement an execution
     * timeout.
     *
     * The interrupt handler can be removed with [[removeInterruptHandler]].
     */
    PureQuickJSVm.prototype.setInterruptHandler = function (cb) {
        var prevInterruptHandler = this.interruptHandler;
        this.interruptHandler = cb;
        if (!prevInterruptHandler) {
            this.ffi.QTS_RuntimeEnableInterruptHandler(this.rt.value);
        }
    };
    /**
     * Remove the interrupt handler, if any.
     * See [[setInterruptHandler]].
     */
    PureQuickJSVm.prototype.removeInterruptHandler = function () {
        if (this.interruptHandler) {
            this.ffi.QTS_RuntimeDisableInterruptHandler(this.rt.value);
            this.interruptHandler = undefined;
        }
    };
    /**
     * Set the max memory this runtime can allocate.
     * To remove the limit, set to `-1`.
     */
    PureQuickJSVm.prototype.setMemoryLimit = function (limitBytes) {
        if (limitBytes < 0 && limitBytes !== -1) {
            throw new Error('Cannot set memory limit to negative number. To unset, pass -1');
        }
        this.ffi.QTS_RuntimeSetMemoryLimit(this.rt.value, limitBytes);
    };
    /**
     * Compute memory usage for this runtime. Returns the result as a handle to a
     * JSValue object. Use [[dump]] to convert to a native object.
     * Calling this method will allocate more memory inside the runtime. The information
     * is accurate as of just before the call to `computeMemoryUsage`.
     * For a human-digestible representation, see [[dumpMemoryUsage]].
     */
    PureQuickJSVm.prototype.computeMemoryUsage = function () {
        return this.memory.heapValueHandle(this.ffi.QTS_RuntimeComputeMemoryUsage(this.rt.value, this.ctx.value));
    };
    /**
     * @returns a human-readable description of memory usage in this runtime.
     * For programmatic access to this information, see [[computeMemoryUsage]].
     */
    PureQuickJSVm.prototype.dumpMemoryUsage = function () {
        return this.ffi.QTS_RuntimeDumpMemoryUsage(this.rt.value);
    };
    return PureQuickJSVm;
}());
exports.PureQuickJSVm = PureQuickJSVm;
/**
 * QuickJSVm wraps a QuickJS Javascript runtime (JSRuntime*) and context (JSContext*).
 * This class's methods return {@link QuickJSHandle}, which wrap C pointers (JSValue*).
 * It's the caller's responsibility to call `.dispose()` on any
 * handles you create to free memory once you're done with the handle.
 *
 * Each QuickJSVm instance is isolated. You cannot share handles between different
 * QuickJSVm instances. You should create separate QuickJSVm instances for
 * untrusted code from different sources for isolation.
 *
 * Use [[QuickJS.createVm]] to create a new QuickJSVm.
 *
 * Create QuickJS values inside the interpreter with methods like
 * [[newNumber]], [[newString]], [[newArray]], [[newObject]],
 * [[newFunction]], and [[newPromise]].
 *
 * Call [[setProp]] or [[defineProp]] to customize objects. Use those methods
 * with [[global]] to expose the values you create to the interior of the
 * interpreter, so they can be used in [[evalCode]].
 *
 * Use [[evalCode]] or [[callFunction]] to execute Javascript inside the VM. If
 * you're using asynchronous code inside the QuickJSVm, you may need to also
 * call [[executePendingJobs]]. Executing code inside the runtime returns a
 * result object representing successful execution or an error. You must dispose
 * of any such results to avoid leaking memory inside the VM.
 *
 * Implement memory and CPU constraints with [[setInterruptHandler]]
 * (called regularly while the interpreter runs) and [[setMemoryLimit]].
 * Use [[computeMemoryUsage]] or [[dumpMemoryUsage]] to guide memory limit
 * tuning.
 */
var QuickJSVm = /** @class */ (function (_super) {
    __extends(QuickJSVm, _super);
    /**
     * Use {@link QuickJS.createVm} to create a QuickJSVm instance.
     */
    function QuickJSVm(args) {
        var _this = _super.call(this, args) || this;
        _this.fnNextId = 0;
        _this.fnMap = new Map();
        /**
         * @hidden
         */
        _this.cToHostCallbackFunction = function (ctx, this_ptr, argc, argv, fn_data) {
            if (ctx !== _this.ctx.value) {
                throw new Error('QuickJSVm instance received C -> JS call with mismatched ctx');
            }
            var fnId = _this.ffi.QTS_GetFloat64(ctx, fn_data);
            var fn = _this.fnMap.get(fnId);
            if (!fn) {
                throw new Error("QuickJSVm had no callback with id ".concat(fnId));
            }
            return lifetime_1.Scope.withScope(function (scope) {
                var thisHandle = scope.manage(new lifetime_1.WeakLifetime(this_ptr, _this.memory.copyJSValue, _this.memory.freeJSValue, _this));
                var argHandles = new Array(argc);
                for (var i = 0; i < argc; i++) {
                    var ptr = _this.ffi.QTS_ArgvGetJSValueConstPointer(argv, i);
                    argHandles[i] = scope.manage(new lifetime_1.WeakLifetime(ptr, _this.memory.copyJSValue, _this.memory.freeJSValue, _this));
                }
                var ownedResultPtr = 0;
                try {
                    var result = fn.apply(thisHandle, argHandles);
                    if (result) {
                        if ('error' in result && result.error) {
                            throw result.error;
                        }
                        var handle = scope.manage(result instanceof lifetime_1.Lifetime ? result : result.value);
                        ownedResultPtr = _this.ffi.QTS_DupValuePointer(_this.ctx.value, handle.value);
                    }
                }
                catch (error) {
                    ownedResultPtr = _this.errorToHandle(error).consume(function (errorHandle) {
                        return _this.ffi.QTS_Throw(_this.ctx.value, errorHandle.value);
                    });
                }
                return ownedResultPtr;
            });
        };
        _this.module = args.module;
        _this.ffi = args.ffi;
        return _this;
    }
    /**
     * `typeof` operator. **Not** [standards compliant](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/typeof).
     *
     * @remarks
     * Does not support BigInt values correctly.
     */
    QuickJSVm.prototype.typeof = function (handle) {
        this.memory.assertOwned(handle);
        return this.ffi.QTS_Typeof(this.ctx.value, handle.value);
    };
    /**
     * Converts `handle` into a Javascript number.
     * @returns `NaN` on error, otherwise a `number`.
     */
    QuickJSVm.prototype.getNumber = function (handle) {
        this.memory.assertOwned(handle);
        return this.ffi.QTS_GetFloat64(this.ctx.value, handle.value);
    };
    /**
     * Converts `handle` to a Javascript string.
     */
    QuickJSVm.prototype.getString = function (handle) {
        this.memory.assertOwned(handle);
        return this.ffi.QTS_GetString(this.ctx.value, handle.value);
    };
    /**
     * Convert a Javascript function into a QuickJS function value.
     * See [[VmFunctionImplementation]] for more details.
     *
     * A [[VmFunctionImplementation]] should not free its arguments or its return
     * value. A VmFunctionImplementation should also not retain any references to
     * its return value.
     *
     * To implement an async function, create a promise with [[newPromise]], then
     * return the deferred promise handle from `deferred.handle` from your
     * function implementation:
     *
     * ```
     * const deferred = vm.newPromise()
     * someNativeAsyncFunction().then(deferred.resolve)
     * return deferred.handle
     * ```
     */
    QuickJSVm.prototype.newFunction = function (name, fn) {
        var fnId = ++this.fnNextId;
        this.fnMap.set(fnId, fn);
        var fnIdHandle = this.newNumber(fnId);
        var funcHandle = this.memory.heapValueHandle(this.ffi.QTS_NewFunction(this.ctx.value, fnIdHandle.value, name));
        // We need to free fnIdHandle's pointer, but not the JSValue, which is retained inside
        // QuickJS for later.
        this.module._free(fnIdHandle.value);
        return funcHandle;
    };
    /**
     * `Promise.resolve(value)`.
     * Convert a handle containing a Promise-like value inside the VM into an
     * actual promise on the host.
     *
     * @remarks
     * You may need to call [[executePendingJobs]] to ensure that the promise is resolved.
     *
     * @param promiseLikeHandle - A handle to a Promise-like value with a `.then(onSuccess, onError)` method.
     */
    QuickJSVm.prototype.resolvePromise = function (promiseLikeHandle) {
        var _this = this;
        this.memory.assertOwned(promiseLikeHandle);
        var vmResolveResult = lifetime_1.Scope.withScope(function (scope) {
            var vmPromise = scope.manage(_this.getProp(_this.global, 'Promise'));
            var vmPromiseResolve = scope.manage(_this.getProp(vmPromise, 'resolve'));
            return _this.callFunction(vmPromiseResolve, vmPromise, promiseLikeHandle);
        });
        if (vmResolveResult.error) {
            return Promise.resolve(vmResolveResult);
        }
        return new Promise(function (resolve) {
            lifetime_1.Scope.withScope(function (scope) {
                var resolveHandle = scope.manage(_this.newFunction('resolve', function (value) {
                    resolve({ value: value && value.dup() });
                }));
                var rejectHandle = scope.manage(_this.newFunction('reject', function (error) {
                    resolve({ error: error && error.dup() });
                }));
                var promiseHandle = scope.manage(vmResolveResult.value);
                var promiseThenHandle = scope.manage(_this.getProp(promiseHandle, 'then'));
                _this.unwrapResult(_this.callFunction(promiseThenHandle, promiseHandle, resolveHandle, rejectHandle)).dispose();
            });
        });
    };
    /**
     * `handle[key]`.
     * Get a property from a JSValue.
     *
     * @param key - The property may be specified as a JSValue handle, or as a
     * Javascript string (which will be converted automatically).
     */
    QuickJSVm.prototype.getProp = function (handle, key) {
        var _this = this;
        this.memory.assertOwned(handle);
        var ptr = this.borrowPropertyKey(key).consume(function (quickJSKey) {
            return _this.ffi.QTS_GetProp(_this.ctx.value, handle.value, quickJSKey.value);
        });
        var result = this.memory.heapValueHandle(ptr);
        return result;
    };
    /**
     * `handle[key] = value`.
     * Set a property on a JSValue.
     *
     * @remarks
     * Note that the QuickJS authors recommend using [[defineProp]] to define new
     * properties.
     *
     * @param key - The property may be specified as a JSValue handle, or as a
     * Javascript string or number (which will be converted automatically to a JSValue).
     */
    QuickJSVm.prototype.setProp = function (handle, key, value) {
        var _this = this;
        this.memory.assertOwned(handle);
        this.borrowPropertyKey(key).consume(function (quickJSKey) {
            return _this.ffi.QTS_SetProp(_this.ctx.value, handle.value, quickJSKey.value, value.value);
        });
        // free newly allocated value if key was a string or number. No-op if string was already
        // a QuickJS handle.
    };
    /**
     * [`Object.defineProperty(handle, key, descriptor)`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/defineProperty).
     *
     * @param key - The property may be specified as a JSValue handle, or as a
     * Javascript string or number (which will be converted automatically to a JSValue).
     */
    QuickJSVm.prototype.defineProp = function (handle, key, descriptor) {
        var _this = this;
        this.memory.assertOwned(handle);
        lifetime_1.Scope.withScope(function (scope) {
            var quickJSKey = scope.manage(_this.borrowPropertyKey(key));
            var value = descriptor.value || _this.undefined;
            var configurable = Boolean(descriptor.configurable);
            var enumerable = Boolean(descriptor.enumerable);
            var hasValue = Boolean(descriptor.value);
            var get = descriptor.get
                ? scope.manage(_this.newFunction(descriptor.get.name, descriptor.get))
                : _this.undefined;
            var set = descriptor.set
                ? scope.manage(_this.newFunction(descriptor.set.name, descriptor.set))
                : _this.undefined;
            _this.ffi.QTS_DefineProp(_this.ctx.value, handle.value, quickJSKey.value, value.value, get.value, set.value, configurable, enumerable, hasValue);
        });
    };
    /**
     * [`func.call(thisVal, ...args)`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Function/call).
     * Call a JSValue as a function.
     *
     * See [[unwrapResult]], which will throw if the function returned an error, or
     * return the result handle directly. If evaluation returned a handle containing
     * a promise, use [[resolvePromise]] to convert it to a native promise and
     * [[executePendingJobs]] to finish evaluating the promise.
     *
     * @returns A result. If the function threw synchronously, `result.error` be a
     * handle to the exception. Otherwise `result.value` will be a handle to the
     * value.
     */
    QuickJSVm.prototype.callFunction = function (func, thisVal) {
        var _this = this;
        var args = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            args[_i - 2] = arguments[_i];
        }
        this.memory.assertOwned(func);
        var resultPtr = this.memory.toPointerArray(args).consume(function (argsArrayPtr) {
            return _this.ffi.QTS_Call(_this.ctx.value, func.value, thisVal.value, args.length, argsArrayPtr.value);
        });
        var errorPtr = this.ffi.QTS_ResolveException(this.ctx.value, resultPtr);
        if (errorPtr) {
            this.ffi.QTS_FreeValuePointer(this.ctx.value, resultPtr);
            return { error: this.memory.heapValueHandle(errorPtr) };
        }
        return { value: this.memory.heapValueHandle(resultPtr) };
    };
    /**
     * Like [`eval(code)`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/eval#Description).
     * Evaluates the Javascript source `code` in the global scope of this VM.
     * When working with async code, you many need to call [[executePendingJobs]]
     * to execute callbacks pending after synchronous evaluation returns.
     *
     * See [[unwrapResult]], which will throw if the function returned an error, or
     * return the result handle directly. If evaluation returned a handle containing
     * a promise, use [[resolvePromise]] to convert it to a native promise and
     * [[executePendingJobs]] to finish evaluating the promise.
     *
     * *Note*: to protect against infinite loops, provide an interrupt handler to
     * [[setInterruptHandler]]. You can use [[shouldInterruptAfterDeadline]] to
     * create a time-based deadline.
     *
     * @returns The last statement's value. If the code threw synchronously,
     * `result.error` will be a handle to the exception. If execution was
     * interrupted, the error will have name `InternalError` and message
     * `interrupted`.
     */
    QuickJSVm.prototype.evalCode = function (code, filename) {
        var _this = this;
        if (filename === void 0) { filename = 'eval.js'; }
        var resultPtr = this.memory.newHeapCharPointer(code).consume(function (charHandle) {
            return _this.ffi.QTS_Eval(_this.ctx.value, charHandle.value, filename);
        });
        var errorPtr = this.ffi.QTS_ResolveException(this.ctx.value, resultPtr);
        if (errorPtr) {
            this.ffi.QTS_FreeValuePointer(this.ctx.value, resultPtr);
            return { error: this.memory.heapValueHandle(errorPtr) };
        }
        return { value: this.memory.heapValueHandle(resultPtr) };
    };
    /**
     * Execute pendingJobs on the VM until `maxJobsToExecute` jobs are executed
     * (default all pendingJobs), the queue is exhausted, or the runtime
     * encounters an exception.
     *
     * In QuickJS, promises and async functions create pendingJobs. These do not execute
     * immediately and need to triggered to run.
     *
     * @param maxJobsToExecute - When negative, run all pending jobs. Otherwise execute
     * at most `maxJobsToExecute` before returning.
     *
     * @return On success, the number of executed jobs. On error, the exception
     * that stopped execution. Note that executePendingJobs will not normally return
     * errors thrown inside async functions or rejected promises. Those errors are
     * available by calling [[resolvePromise]] on the promise handle returned by
     * the async function.
     */
    QuickJSVm.prototype.executePendingJobs = function (maxJobsToExecute) {
        if (maxJobsToExecute === void 0) { maxJobsToExecute = -1; }
        var resultValue = this.memory.heapValueHandle(this.ffi.QTS_ExecutePendingJob(this.rt.value, maxJobsToExecute));
        var typeOfRet = this.typeof(resultValue);
        if (typeOfRet === 'number') {
            var executedJobs = this.getNumber(resultValue);
            resultValue.dispose();
            return { value: executedJobs };
        }
        else {
            return { error: resultValue };
        }
    };
    // customizations
    /**
     * Dump a JSValue to Javascript in a best-effort fashion.
     * Returns `handle.toString()` if it cannot be serialized to JSON.
     */
    QuickJSVm.prototype.dump = function (handle) {
        this.memory.assertOwned(handle);
        var type = this.typeof(handle);
        if (type === 'string') {
            return this.getString(handle);
        }
        else if (type === 'number') {
            return this.getNumber(handle);
        }
        else if (type === 'undefined') {
            return undefined;
        }
        var str = this.ffi.QTS_Dump(this.ctx.value, handle.value);
        try {
            return JSON.parse(str);
        }
        catch (err) {
            return str;
        }
    };
    /**
     * Unwrap a SuccessOrFail result such as a [[VmCallResult]] or a
     * [[ExecutePendingJobsResult]], where the fail branch contains a handle to a QuickJS error value.
     * If the result is a success, returns the value.
     * If the result is an error, converts the error to a native object and throws the error.
     */
    QuickJSVm.prototype.unwrapResult = function (result) {
        var _this = this;
        if (result.error) {
            var dumped = result.error.consume(function (error) { return _this.dump(error); });
            if (dumped && typeof dumped === 'object' && typeof dumped.message === 'string') {
                var exception = new Error(dumped.message);
                if (typeof dumped.name === 'string') {
                    exception.name = dumped.name;
                }
                throw exception;
            }
            throw dumped;
        }
        return result.value;
    };
    QuickJSVm.prototype.borrowPropertyKey = function (key) {
        if (typeof key === 'number') {
            return this.newNumber(key);
        }
        if (typeof key === 'string') {
            return this.newString(key);
        }
        // key is already a JSValue, but we're borrowing it. Return a static handle
        // for internal use only.
        return new lifetime_1.StaticLifetime(key.value, this);
    };
    QuickJSVm.prototype.errorToHandle = function (error) {
        var _this = this;
        if (error instanceof lifetime_1.Lifetime) {
            return error;
        }
        var errorHandle = this.memory.heapValueHandle(this.ffi.QTS_NewError(this.ctx.value));
        if (error.name !== undefined) {
            this.newString(error.name).consume(function (handle) { return _this.setProp(errorHandle, 'name', handle); });
        }
        if (error.message !== undefined) {
            this.newString(error.message).consume(function (handle) { return _this.setProp(errorHandle, 'message', handle); });
        }
        // Disabled due to security leak concerns
        if (error.stack !== undefined) {
            //const handle = this.newString(error.stack)
            // Set to fullStack...? For debugging.
            //this.setProp(errorHandle, 'fullStack', handle)
            //handle.dispose()
        }
        return errorHandle;
    };
    return QuickJSVm;
}(PureQuickJSVm));
exports.QuickJSVm = QuickJSVm;
//# sourceMappingURL=vm.js.map